#include "vector_field.h"
#include "vecteur_nD.h"

// Créé un champ de vecteur à partir des expressions qui le composent.
vector_field::vector_field()
{
};

// Créé un champ de vecteur à partir d'un string.
// ensuite on remplit le expr2 en substituant 1 à partial_x et 0 à partial_y, etc ...
/* vector_field::vector_field(string a)
{
  expression exp(a);
  for (set<string>::const_iterator i = exp.symboles.begin(); i != exp.symboles.end(); ++i)
    {
      expression ei = exp.substitue_partial(*i);
      expr2[*i] = ei;
    }
};
*/

// expressions du champ de vecteur vers sortie standard
void vector_field::out()
{
  cout << ""  << endl;
  for (map<string,expression>::const_iterator i = expr2.begin(); i!=expr2.end(); ++i)
    { 
      cout << i->first << ":" << i->second.chaine <<endl ;
    };
};


// génère l'équation différentielle satisfaite par la différentielle du flot
// l'équation importante est la suivante : \dot{J} = Df . J, ce qui se réécrit pour chaque coordonnée :
//\dot{ax_ix_j} = \sum_k (  \frac{\partial f_x_i}{\partial x_k} * ax_kx_j ) 
vector_field vector_field::d()
{
  //  vector_field result = (*this);
  vector_field result;
  for (map<string,expression>::const_iterator j = expr2.begin(); j!=expr2.end(); ++j)
    { 
        for (map<string,expression>::const_iterator i = expr2.begin(); i!=expr2.end(); ++i)
	  { 
	    // génération de l'équation de x_i dans la direction j
	    string x_i =  i->first;
	    string x_j =  j->first;

	    string factor_x_ix_j; 
	    bool premier = true;
	    
	    for (map<string,expression>::const_iterator k = expr2.begin(); k!=expr2.end(); ++k)
	      { 
		string x_k =  k->first;
		expression la = i->second;
		string dfx_idx_k = la.derive(x_k).chaine ;
		if (premier) 
		  {
		    factor_x_ix_j = "(" + dfx_idx_k + ")*a" + x_k + x_j ;
		    premier = false;
		  }
		else 
		  factor_x_ix_j = factor_x_ix_j + "+(" + dfx_idx_k + ")*a" + x_k + x_j ;
	      }
	    expression efactor_x_ix_j  ( factor_x_ix_j );
	    efactor_x_ix_j.Simplify();
	    result.expr2["a" + x_i + x_j] = efactor_x_ix_j;
	    // cout << compte++ << ":a" + x_i + x_j << ":" << efactor_x_ix_j.chaine << endl;
	  } 
    };
   return result;
};


// genere un domain qui contient la matrice identité
domain vector_field::identite()
{
  domain result;
  for (map<string,expression>::const_iterator j = expr2.begin(); j!=expr2.end(); ++j)
    { 
        for (map<string,expression>::const_iterator i = expr2.begin(); i!=expr2.end(); ++i)
	  { 
	    // génération de l'équation de x_i dans la direction j
	    string x_i =  i->first;
	    string x_j =  j->first;
	    if (x_i == x_j)
	      {result["a" + x_i + x_j] = interval(1,1);}
	    else 
	      {result["a" + x_i + x_j] = interval(0,0);}
	  } 
    };
   return result;
};


// renvoie une borne supérieure de la norme de $f$ sur le domaine x
double vector_field::max_norme_f(domain x)
{
  bool premier = true;
  expression result;
  for (map<string,expression>::const_iterator i = expr2.begin(); i!=expr2.end(); ++i)
    { 
      expression e_t = i->second;
      if (premier) 
	{
	  premier = false;
	  result = e_t.pow(2);
	}
      else 
	{
	  result = result + e_t.pow(2);
	}
    }
  //  result.ToLatex("essai2.tex");
  return sqrt(result.Evalue(x).sup());
};


// renvoie la valeur d'un champ de vecteurs sur un domaine
domain vector_field::f(domain x)
{
  domain result;
  for (map<string,expression>::const_iterator i = expr2.begin(); i!=expr2.end(); ++i)
    { 
      expression e_t = i->second;
      result[i->first] = e_t.Evalue(x);
    } 
  return result;
};


// renvoie la valeur x + h*f(x) en utilisant la forme centrée
domain vector_field::f0_1(domain x , double h)
{
  std::ostringstream oss;
  oss << h;
  std::string H = oss.str();
  expression hs(H);
  domain result;
  for (map<string,expression>::const_iterator i = expr2.begin(); i!=expr2.end(); ++i)
    { 
      expression f0_i (i->first);
      expression f0_1 = f0_i + hs*(i->second);
      result[i->first] =  f0_1.forme_centre(x);
    };
  return result;
};


// renvoie la valeur 
domain vector_field::f2(domain x)
{
  domain result;
  for (map<string,expression>::const_iterator i = expr2.begin(); i!=expr2.end(); ++i)
    { 
      expression f2_i ("0");
        for (map<string,expression>::const_iterator j = expr2.begin(); j!=expr2.end(); ++j)
	  { 
	    string variable = (j->first);
	    expression e_t = i->second;
	    expression e_ti = e_t.derive(variable);
      	    f2_i = f2_i + e_ti*(j->second);
	  } 
	result[i->first] =  f2_i.Evalue(x);
    };
  return result;
};



// guaranteed integration
// ici T est un grand intervalle de temps, l'idée est d'intégrer sur les petits pas de temps dt 
// jusqu'à ce que la somme des dt égale T.
// On fait bien attention d'avoir un dt qui satisfait l'inclusion qui suffit à appliquer Picard. 
domain vector_field::integrate(domain x, double T)
{
  curve result;
  domain result_t = x;
  double t = 0;
  result.push_back(result_t);
  do
    {
      double maximum = max_norme_f(result_t);
      double h = min(0.5*maximum,T-t);      
      while (!(validation(result_t,h)))
	{
  	  h  = 0.5*h;
	}
      t = t + h;
      result_t =  solve_taylor(result_t,h,validated);
      result.push_back(result_t);
    }
  while ( t < T);
  return result[result.size()-1];
};

// Ici on découpe le temps T de facon uniforme en nb dt et on appelle la fonction integrate 
curve vector_field::integrate(domain x, double T, int nb)
{
  curve result;
  double t = 0;
  x["t"] = interval(t,t);
  result.push_back(x);
  double h = T / nb;
  for (int i=0; i< nb; i++ )
    {   
      domain te = integrate(result[result.size()-1],h); 
      t = t +h;
      te["t"] = interval(t,t); 
      result.push_back(te);
      te.out();
    }
  return result;
}


vector_field vector_field::direct_sum(vector_field v)
{	
  vector_field result = (*this);
  for (map<string,expression>::const_iterator i = v.expr2.begin(); i!=v.expr2.end();++i)
    result.expr2[i->first] = v.expr2[i->first];
  return result;
}



curve vector_field::monotone_integrate(domain x, double T, int nb)
{

  vector_field derivative_temp = (*this).d();
  vector_field derivative = derivative_temp.direct_sum((*this));

  derivative.out();

  curve result;
  double t = 0;
  x["t"] = interval(t,t);
  domain x_Id = x.direct_sum(identite());
  result.push_back(x_Id);

  double h = T / nb;
  for (int i=0; i< nb; i++ )
    { 

      domain a_integrer = result[result.size()-1];
      domain a_integrer_sauve = a_integrer;
      
      a_integrer["axx"] = 1;
      a_integrer["ayy"] = 1;
      a_integrer["axy"] = 0;
      a_integrer["ayx"] = 0;
      
      domain te ;
      
      te = derivative.integrate(a_integrer,h);
      //te.out();
      t = t + h;

      domain b;
      b["x"] = interval(x["x"].inf(),x["x"].inf());
      b["y"] = interval(x["y"].inf(),x["y"].inf());
      //      b = b.direct_sum(identite());
      curve tb = (*this).integrate(b,t,i);
      b = tb[tb.size()-1];
      curveb.push_back(b);


      domain c;
      c["x"] = interval(x["x"].sup(),x["x"].sup());
      c["y"] = interval(x["y"].inf(),x["y"].inf());
      //      c = c.direct_sum(identite());
      curve tc = (*this).integrate(c,t,i);
      c = tc[tc.size()-1];
      curvec.push_back(c);
      
      domain d;
      d["x"] = interval(x["x"].sup(),x["x"].sup());
      d["y"] = interval(x["y"].sup(),x["y"].sup());
      //d = d.direct_sum(identite());
      curve td = (*this).integrate(d,t,i);
      d = td[td.size()-1];
      curved.push_back(d);

      domain e;
      e["x"] = interval(x["x"].inf(),x["x"].inf());
      e["y"] = interval(x["y"].sup(),x["y"].sup());
      //e = e.direct_sum(identite());
      curve tef = (*this).integrate(e,t,i);
      e = tef[tef.size()-1];	
      curvee.push_back(e);
		

      domain enveloppe = b.hull(c);
      enveloppe = enveloppe.hull(d);
      enveloppe = enveloppe.hull(e);

      //enveloppe = enveloppe.direct_sum(identite());

      te["x"] = enveloppe["x"];
      te["y"] = enveloppe["y"];
     
      domain nouveau_te;

      nouveau_te["axx"] = a_integrer_sauve["axx"]*te["axx"]+a_integrer_sauve["axy"]*te["ayx"];
      nouveau_te["axy"] = a_integrer_sauve["axx"]*te["axy"]+a_integrer_sauve["axy"]*te["ayy"];
      nouveau_te["ayx"] = a_integrer_sauve["ayx"]*te["axx"]+a_integrer_sauve["ayy"]*te["ayx"];
      nouveau_te["ayy"] = a_integrer_sauve["ayx"]*te["axy"]+a_integrer_sauve["ayy"]*te["ayy"];

      te["axx"] = nouveau_te["axx"];
      te["axy"] = nouveau_te["axy"];
      te["ayx"] = nouveau_te["ayx"];
      te["ayy"] = nouveau_te["ayy"];

      // te = enveloppe;
      te["t"] = interval(t,t); 

      te.out();
      
      result.push_back(te);
    }
  
  
  
  return result;
}



// ----------------------------------------------------------------------------------------------------
// Méthode d'intégration numérique garantie.
// basée sur l'opérateur de Picard
// ----------------------------------------------------------------------------------------------------
domain vector_field::solve_taylor(domain x, double h, domain valide)
{
  domain result;
  interval H = (h*h/2);
  result = f0_1(x,h) + f2(valide)*H;
  return result;
};


// methode basée sur la différentiation automatique
taylor vector_field::Taylor(domain x, int ordre )
{
  taylor result(x);
  for (int j=0; j<ordre ; j++)
    {
      for (map<string,expression>::const_iterator i = expr2.begin(); i!=expr2.end(); ++i)
	{ 
	  expression e_t = i->second;
	  vector <interval> temp = e_t.taylor_ode(result,j);
	  result[i->first].push_back(temp[j] / interval(j+1)); // la division par j+1 s'explique par le fait que Taylor( dot x ) = shift [ x . (0 1 2 3 4 5 6 ... )] (où . est le produit ponctuel)
	};
  };

  return result;
  
}


// ----------------------------------------------------------------------------------------------------
// Méthode de validation numérique.
// ----------------------------------------------------------------------------------------------------
bool vector_field::validation(domain d, double h)
{
 domain result;
 result = d + f(d)*interval(h,h);
 result = d.hull(result);

 double eps = h;
 bool valide = false;
 int n=0;
 do 
   {   
     domain temp = d + f(result)*interval(0,h);
     valide = temp.subset(result);
     if (valide==true) 
       {
	 validated = result;
	 return valide;
       }
     result = result.hull(result + f(result)*eps);
     n++;
   }
 while ((valide!=true) && (n<10));
 return valide;
};



// ----------------------------------------------------------------------------------------------------
// Méthode d'intégration numérique non garantie à un pas. Euler.
// ----------------------------------------------------------------------------------------------------
domain vector_field::Next_Euler(double h, domain point)
{
return point + f(point)*h;
}
;

curve vector_field::Phase_Curve_via_Euler(double h, domain point, double fin)
{
  domain X_h = point;
  curve result;
  double temps;
  do
    {    
      result.push_back(X_h);
      temps = temps + h;
      X_h = Next_Euler(h, X_h);
    } 
  while (temps <= fin);
  return result;
};


